from marvin.core.suite import Suite
from marvin.core.step import Step
from marvin.core.test_script import TestScript
from marvin.version import __version__


__all__ = ["Suite", "TestScript", "Step", "__version__"]



import sys

from marvin.data.data_providers.null_data_provider import NullDataProvider
from marvin.core.status import Status
from marvin.exceptions import ContextSkippedException
from marvin.report.events import TestStartedEvent, TestEndedEvent, TestSetupStartedEvent, TestSetupEndedEvent, \
    TestIterationStartedEvent, TestIterationEndedEvent, TestTearDownStartedEvent, TestTearDownEndedEvent
from marvin.util import NO_EXCEPTION


class TestRunner(object):
    """
    Internal Marvin Test executor. Implement the logic of a test execution flow, signals, events, etc
    """

    def __init__(self, test_script, data_provider):
        self._test = test_script
        self._data_provider = data_provider or NullDataProvider()
        self._status = None
        self._exceptions = []
        self._skip_iteration = False
        self._skip_teardown = False

    def execute(self):
        self._test_meta_override()
        if not self._should_run_test():
            return
        test_started = TestStartedEvent(self._test, self._data_provider)
        self._test.publisher.notify(test_started)

        self._execute()

        test_ended = TestEndedEvent(self._test, self._data_provider,
                                    test_started.timestamp, self._status, self._exceptions)
        self._test.publisher.notify(test_ended)
        self._test.ctx.sub_context_finished(self._status)

    def _execute(self):
        self._run_phase('setup', self._data_provider.setup_data, TestSetupStartedEvent, TestSetupEndedEvent)

        for iteration in self._data_provider.iterations:
            if self._test.ctx.tags_match(self._iteration_tags(iteration)):
                self._run_phase('run', iteration, TestIterationStartedEvent, TestIterationEndedEvent)

        self._run_phase('tear_down', self._data_provider.tear_down_data,
                        TestTearDownStartedEvent, TestTearDownEndedEvent)

    def _run_phase(self, phase_type, data, started_event_class, ended_event_class):
        if self._should_skip_phase(phase_type):
            return

        status = Status.PASS
        exception = NO_EXCEPTION
        phase_started = started_event_class(self._test, self._data_provider, data)
        self._test.publisher.notify(phase_started)
        try:
            phase_data = data.data if phase_type == 'run' else data
            getattr(self._test, phase_type)(phase_data)
        except ContextSkippedException:
            status = Status.SKIP
            exception = sys.exc_info()
        except:
            status = Status.FAIL
            exception = sys.exc_info()
        finally:
            self._test.publisher.notify(
                ended_event_class(self._test, self._data_provider, data, phase_started.timestamp, status, exception)
            )
            if exception != NO_EXCEPTION:
                self._exceptions.append(exception)
            self._report_phase_status(phase_type, status)

    def _should_run_test(self):
        return any(self._test.ctx.tags_match(self._iteration_tags(iteration))
                   for iteration in self._data_provider.iterations)

    def _iteration_tags(self, iteration):
        return self._test.tags | (iteration.tags or set())

    def _report_phase_status(self, phase_type, status):
        if phase_type == 'setup' and status in [Status.FAIL, Status.SKIP]:
            self._skip_iteration = True
        if phase_type == 'setup' and status == Status.SKIP:
            self._skip_teardown = True
        if phase_type == 'setup' or status == Status.FAIL:
            self._status = status

    def _should_skip_phase(self, phase_type):
        if phase_type == 'tear_down':
            return self._skip_teardown
        if phase_type == 'run':
            return self._skip_iteration
        return False

    def _test_meta_override(self):
        if self._data_provider.name:
            self._test.name = self._data_provider.name
        if self._data_provider.description:
            self._test.description = self._data_provider.description
        tags = self._data_provider.tags
        if isinstance(tags, set):
            self._test.tag(*tags)



import os
import glob

from marvin import Suite, TestScript
from marvin.util.files import ClassLoader, FileFinder
from marvin.util import compat
from marvin.data import DataProviderRegistry


def default_config():
    return {
        'tests_path': '.',
        'filter': {
            'with_tags': [],
            'without_tags': []
        },
        'hook_module': 'marvin_conf.py',
        'env': None,
        'environments': {}
    }


class RuntimeSuite(Suite):
    def __init__(self, config_file=None, tests_path=None, with_tags=None, without_tags=None):
        super(RuntimeSuite, self).__init__()
        self._supported_data_file_extensions = DataProviderRegistry.supported_file_extensions()
        self._load_marvin_config(config_file=config_file,
                                 tests_path=tests_path,
                                 with_tags=with_tags,
                                 without_tags=without_tags)
        self._load_test_environment_config()
        self._load_hook_module()
        self._root_dir = self.cfg.marvin.get('tests_path', '.')
        self._tags_matcher = TagsMatcher(*self._get_tag_filters())

    # Override
    def tests(self):
        test_script_class_loader = ClassLoader(TestScript)

        python_file_finder = FileFinder(self._root_dir,
                                        filename_matcher=lambda name: name.lower().endswith('.py'))

        for python_file in python_file_finder.find_all():
            directory, filename = os.path.split(python_file)
            filename_base, _ = os.path.splitext(filename)

            mod = compat.import_module(python_file, module_name=filename_base)

            test_script_class = test_script_class_loader.find(mod)
            if not test_script_class:
                continue

            if self._tags_matcher.blacklisted(test_script_class.class_tags()):
                continue

            test_script_id = filename_base
            for data_provider in self._find_data_for_test_script(test_script_id, directory):
                yield test_script_class, data_provider

    # Override
    def tags_match(self, tags):
        return self._tags_matcher.matches(tags)

    def _find_data_for_test_script(self, test_script_id, test_script_directory):
        data_file_matcher = DataFileMatcher(test_script_id, self._supported_data_file_extensions)
        data_file_finder = FileFinder(test_script_directory, filename_matcher=data_file_matcher)

        data_found = False
        for data_file in data_file_finder.find_all():
            provider = DataProviderRegistry.data_provider_for(data_file)

            if not provider or self._tags_matcher.blacklisted(provider.tags):
                continue

            data_found = True
            yield provider

        if not data_found:
            yield DataProviderRegistry.data_provider_for(None)

    def _load_marvin_config(self, **options):
        self.cfg.set('marvin', default_config())

        cfg_file = (options.get('config_file') or os.environ.get('MARVIN_CONFIG') or
                    next((path for path in ['marvin.yaml', 'marvin.yml', 'marvin.json'] if os.path.isfile(path)), None))
        if cfg_file:
            self.cfg.load_into('marvin', cfg_file)

        cli_overrides = {}
        if options.get('tests_path'):
            cli_overrides['tests_path'] = options['tests_path']
        if options.get('with_tags'):
            cli_overrides.setdefault('filter', {})['with_tags'] = options['with_tags']
        if options.get('without_tags'):
            cli_overrides.setdefault('filter', {})['without_tags'] = options['without_tags']

        self.cfg.set('marvin', cli_overrides)

    def _load_test_environment_config(self):
        env = os.environ.get('MARVIN_ENV', self.cfg.marvin.get('env'))
        env_files_globs = self.cfg.marvin.get('environments', {}).get(env, [])
        for glob_matcher in env_files_globs:
            self.cfg.load(*glob.glob(glob_matcher))

    def _load_hook_module(self):
        file_name = self.cfg.marvin.get('hook_module')
        if not (file_name and os.path.isfile(file_name)):
            return
        mod = compat.import_module(file_name, module_name='marvin_hook')
        if hasattr(mod, 'main') and callable(mod.main):
            mod.main(self.publisher, self.cfg)

    def _get_tag_filters(self):
        filter_config = self.cfg.marvin.get('filter', {})
        with_tags = set(filter_config.get('with_tags', []))
        without_tags = set(filter_config.get('without_tags', []))
        return with_tags, without_tags


class TagsMatcher(object):
    def __init__(self, with_tags, without_tags):
        self._with = with_tags
        self._without = without_tags

    def blacklisted(self, tags):
        tags = tags or set()
        return any(tags & self._without)

    def whitelisted(self, tags):
        tags = tags or set()
        return not self._with or (tags & self._with)

    def matches(self, tags):
        return self.whitelisted(tags) and not self.blacklisted(tags)


class DataFileMatcher(object):
    """
    Callable that returns True if the given filename starts with
    the the test script id, and has a supported data extension
    """
    def __init__(self, test_script_id, supported_data_file_extensions):
        self._test_script_id = test_script_id
        self._extensions = supported_data_file_extensions

    def __call__(self, filename):
        parts = filename.split('.')
        return (len(parts) > 1
                and parts[0] == self._test_script_id
                and parts[-1] in self._extensions)



from marvin import Step

import requests
import bs4


class CheckWebsiteTitle(Step):
    """Checks if a website has an expected title"""

    NAME = "Check Website Title"

    def run(self, website, expected_title):

        if expected_title == "Google":
            self.skip("Skipping test")

        response = requests.get(website)
        html = bs4.BeautifulSoup(response.text, "html.parser")
        real_title = html.title.text

        assert real_title == expected_title,\
            "Title '%s' is not equal to expected '%s'" % (expected_title, real_title)

        return real_title



import sys
import argparse

import marvin
from marvin.runner.runner import Runner


def main(exit_fn=sys.exit, args=None):
    options = parse(args)
    if options.version:
        sys_info()
    else:
        exit_fn(Runner(options).run())


def parse(args):
    parser = argparse.ArgumentParser(description='Marvin test runner')
    parser.add_argument('tests_path', metavar='DIRECTORY', nargs='?', default=None,
                        help='root directory containing tests')
    parser.add_argument('--config', '-c', dest='config',
                        help='path to config file')
    parser.add_argument('--tags', nargs='*', dest='with_tags',
                        help='only run tests containing any of the given tags')
    parser.add_argument('--no-tags', nargs='*', dest='without_tags',
                        help='don\'t run tests containing any of the given tags')
    parser.add_argument('--version', action='store_true',
                        help='display Marvin version and exit')

    return parser.parse_args(args=args)


def sys_info():
    print("Marvin: %s" % marvin.__version__)
    print("Python: %s" % sys.version.replace('\n', ' - '))
    print("Platform: %s" % sys.platform)




import inspect
import sys

from marvin.core.status import Status
from marvin.exceptions import ContextSkippedException, ExpectedExceptionNotRaised, StepsFailedInContext
from marvin.report.events import StepStartedEvent, StepEndedEvent, StepSkippedEvent
from marvin.util import compat, NO_EXCEPTION


class StepRunner(object):
    """
    Internal Marvin Step executor. Implement the logic of a step execution flow, signals, events, etc
    """

    def __init__(self, step, args, kwargs):
        self._step = step
        self._args = list(args)  # make args mutable
        self._kwargs = kwargs
        self._execution = None
        self._start_time = None
        self._result = Result(None)
        self._step_exception = NO_EXCEPTION
        self._context_exception = NO_EXCEPTION

    @property
    def result(self):
        return self._result.get()

    def __enter__(self):
        start_event = StepStartedEvent(self._step, self._args, self._kwargs)
        self._start_time = start_event.timestamp
        self._step.publisher.notify(start_event)

        result, self._step_exception = self._do_run()
        self._result.set(result)

        is_exception = self._step_exception != NO_EXCEPTION
        expected = is_exception and self._is_exception_expected(self._step_exception, self._step.expected_exceptions)

        # Decide whether we should yield the result or fail right away
        if is_exception and self._step.safe_exec:
            # TODO: This is more a language constraint than a feature.
            # Alternatively we could make `safely` and `do` to be incompatible.
            # Or find a way to inject an exception to be raised immediately inside the context
            return self._step, self._step_exception

        elif is_exception and not expected:
            self.__exit__(*NO_EXCEPTION)

        return self._step, self.result

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._context_exception = (exc_type, exc_val, exc_tb)
        sub_context_failures = self._any_sub_steps_failed()

        if self._step_exception != NO_EXCEPTION:
            exception = self._step_exception
            status, to_raise = self._when_exception(self._step_exception)
        elif self._context_exception != NO_EXCEPTION:
            exception = self._context_exception
            status, to_raise = self._when_exception(self._context_exception)
        elif sub_context_failures != NO_EXCEPTION:
            exception = sub_context_failures
            status, to_raise = self._when_exception(sub_context_failures)
        else:
            status, to_raise = self._when_no_exception()
            exception = to_raise

        skip_exception = None
        if self._step_exception[0] == ContextSkippedException:
            skip_exception = self._step_exception

        elif self._context_exception[0] == ContextSkippedException:
            skip_exception = self._context_exception

        if skip_exception:
            status = Status.SKIP
            skip_event = StepSkippedEvent(self._step, skip_exception)
            self._step.publisher.notify(skip_event)

        if self._step.shall_pass and status == Status.FAIL:
            status = Status.PASS

        end_event = StepEndedEvent(self._step,
                                   status,
                                   self._result,
                                   self._start_time,
                                   exception)

        self._step.publisher.notify(end_event)

        # Notify the result to the parent context
        self._step.ctx.sub_context_finished(status)

        if to_raise[1] and not self._step.safe_exec:
            raise compat.raise_exc_info(*to_raise)

        return True

    def _do_run(self):
        result = None
        exception = NO_EXCEPTION
        try:
            result = self._step.run(*self._args, **self._kwargs)
        except Exception:
            exception = sys.exc_info()

        return result, exception

    def _any_sub_steps_failed(self):
        failed_subcontexts = self._step.context_summary[Status.FAIL]
        if failed_subcontexts:
            return StepsFailedInContext, StepsFailedInContext(failed_subcontexts), None
        return NO_EXCEPTION

    def _when_no_exception(self):
        if not self._step.expected_exceptions:
            return Status.PASS, NO_EXCEPTION

        to_raise = (ExpectedExceptionNotRaised, ExpectedExceptionNotRaised(self._step.expected_exceptions), None)
        return Status.FAIL, to_raise

    def _when_exception(self, exception):
        was_expected = self._is_exception_expected(exception, self._step.expected_exceptions)
        if self._step.shall_pass or was_expected:
            status = Status.PASS
        else:
            status = Status.FAIL
        if self._step.safe_exec or was_expected:
            to_raise = NO_EXCEPTION
        else:
            to_raise = exception
        return status, to_raise

    def _is_exception_expected(self, exception, expectation):
        exc_type, exc_val, exc_tb = exception
        if isinstance(expectation, list):
            return any(self._is_exception_expected(exception, expect) for expect in expectation)
        if inspect.isclass(expectation):
            return issubclass(expectation, exc_type)
        if callable(expectation):
            return expectation(exc_val)
        return False


class Result(object):
    """
    Wraps a step result so plugins can change its value even if type is immutable
    """
    def __init__(self, result):
        self._result = result

    def get(self):
        """Get the step result object"""
        return self._result

    def set(self, value):
        """Change the step result object"""
        self._result = value




class Reportable(object):
    """
    Mix-In class for all those Marvin entities that are considered to be 'reportable'
    E.g. Steps, TestScripts, and Suites
    It adds capabilities to define name, description, and tags properties
    """

    def __init__(self):
        self._runtime_tags = set()
        self._removed_tags = set()
        self._name = getattr(self, 'NAME', self.__class__.__name__)
        self._description = getattr(self, 'DESCRIPTION', self._default_description())

    @property
    def name(self):
        """Name property for this instance"""
        return self._name

    @name.setter
    def name(self, new_name):
        self._name = str(new_name)

    @property
    def description(self):
        """Description property for this instance"""
        return self._description

    @description.setter
    def description(self, new_description):
        self._description = str(new_description)

    @property
    def tags(self):
        """Retrieve the full set of tags for this instance (class and instance tags)"""
        hierarchy_tags = self.class_tags()
        return (self._runtime_tags | hierarchy_tags) - self._removed_tags

    def tag(self, *tags):
        """Dynamically add tags to this instance"""
        self._runtime_tags |= set(str(tag) for tag in tags)

    def untag(self, *tags):
        """Remove tags (instance or class) for this instance"""
        self._removed_tags |= set(str(tag) for tag in tags)

    @classmethod
    def class_tags(cls):
        """Retrieve all the tags of this class and its superclasses"""
        tags = set(getattr(cls, 'TAGS', []))
        for base_class in cls.__bases__:
            if issubclass(base_class, Reportable) and base_class != Reportable:
                tags |= base_class.class_tags()
                break
        return tags

    def _default_description(self):
        if self.__class__.__doc__:
            return self.__class__.__doc__.strip()
        return 'No description'




from marvin.core.context import Context
from marvin.core.reportable import Reportable
from marvin.core.status import Status
from marvin.core.test_running_context import TestRunningContext
from marvin.report.events import SuiteStartedEvent, SuiteEndedEvent


class Suite(Context, TestRunningContext, Reportable):
    """
    A context representing a collection of tests
    """

    def __init__(self):
        Context.__init__(self, parent_context=None)
        TestRunningContext.__init__(self)
        Reportable.__init__(self)

    def execute(self):
        """
        Executes the tests in this suite
        """
        start_event = SuiteStartedEvent(self)
        self.publisher.notify(start_event)

        status = self._execute()

        self.publisher.notify(SuiteEndedEvent(self, start_event.timestamp, status))
        return status

    def tests(self):
        """To be redefined by specific Suite implementations. The method must return
        an iterable object (e.g. a list or a generator) of tuples (."""

        raise NotImplementedError("Method run must be redefined")

    def _execute(self):
        # TODO: subclasses should be able to define how tests are executed (e.g. allowing parallel execution)
        for test_class, data_provider in self.tests():
            self.test(test_class).execute(data_provider=data_provider)

        if self.context_summary[Status.FAIL]:
            return Status.FAIL
        elif self.context_summary[Status.PASS]:
            return Status.PASS
        return Status.SKIP



class IterationData(object):
    """
    Models an iteration data and meta data.
    """

    def __init__(self, data=None, name=None, description=None, tags=None, **_kwargs):
        self._data = data
        self._name = name
        self._description = description
        self._tags = set(str(tag) for tag in tags) if tags else None

    @property
    def name(self):
        """
        iteration name
        :return: [None|String]
        """
        return self._name

    @property
    def description(self):
        """
        iteration description
        :return: [None|String]
        """
        return self._description

    @property
    def tags(self):
        """
        iteration tags
        :return: set
        """
        return self._tags

    @property
    def data(self):
        """
        iteration data
        :return: anything
        """
        return self._data



import sys

IS_PYTHON_3 = sys.version_info >= (3, 0)
IS_PYTHON_2 = sys.version_info[0] == 2


def raise_exc_info(_exc_type, exc_val, exc_tb):
    if IS_PYTHON_3:
        return exc_val.with_traceback(exc_tb)
    return exc_val, None, exc_tb


def import_module(module_path, module_name):
    if sys.version_info >= (3, 5):
        import importlib.util
        spec = importlib.util.spec_from_file_location(module_name, module_path)
        mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mod)

    elif sys.version_info >= (3, 3):
        from importlib.machinery import SourceFileLoader
        mod = SourceFileLoader(module_name, module_path).load_module()

    elif IS_PYTHON_2:
        import imp
        mod = imp.load_source(module_name, module_path)
    else:
        raise RuntimeError("Python version not supported: %s" % (".".join(sys.version_info)))

    return mod


def string_io(*args, **kwargs):
    if IS_PYTHON_3:
        from io import StringIO
    else:
        from StringIO import StringIO
    return StringIO(*args, **kwargs)


def urllib_mod():
    if IS_PYTHON_3:
        import urllib.request
        return urllib.request
    else:
        import urllib2
        return urllib2



from marvin.core.status import Status
from marvin.exceptions import ContextSkippedException
from marvin.report import Publisher
from marvin.util.config import Config


class Context(object):

    def __init__(self, parent_context=None):
        self._ctx = parent_context
        self._publisher = self.ctx.publisher if self.ctx else Publisher()
        self._cfg = self.ctx.cfg if self.ctx else Config()
        self._context_summary = {
            Status.PASS: 0,
            Status.FAIL: 0,
            Status.SKIP: 0
        }

    @property
    def ctx(self):
        """This context's context (i.e. the parent context)"""
        return self._ctx

    @property
    def cfg(self):
        """This context's config object"""
        return self._cfg

    @property
    def publisher(self):
        """The publisher instance for this context"""
        return self._publisher

    @property
    def level(self):
        """Depth level of this context (i.e. how many ancestors contexts it has)"""
        if not self._ctx:
            return 0
        return self._ctx.level + 1

    @property
    def context_summary(self):
        """Result summary of direct sub contexts executed within this context"""
        return self._context_summary

    def skip(self, reason=None):
        """Trigger a signal for this context execution to be skipped"""
        raise ContextSkippedException(self, reason=reason)

    def sub_context_finished(self, status):
        """Keep count of sub-context results within this context"""
        self.context_summary[status] += 1



import sys

from marvin.core.status import Status
from marvin.runner.runtime_suite import RuntimeSuite
from marvin.report.observers.event_logger import EventLogger


class Runner(object):
    def __init__(self, options):
        self._setup_env()
        self._suite = RuntimeSuite(config_file=options.config, tests_path=options.tests_path,
                                   with_tags=options.with_tags, without_tags=options.without_tags)
        self._load_observers()

    def run(self):
        status = self._suite.execute()
        return 0 if status in [Status.PASS, Status.SKIP] else 1

    def _setup_env(self):
        sys.path.insert(1, '.')

    def _load_observers(self):
        EventLogger(self._suite.publisher)



from marvin.core.context import Context


class StepRunningContext(object):
    """
    A Mix-In for contexts that are allowed to run steps.
    E.g. TestScripts, and Steps contexts can run steps
    """

    def __init__(self):
        assert isinstance(self, Context)

    def step(self, step_class):
        """Instantiates a new step with this as it's parent context"""
        # TODO: assert step_class is subclass of Step
        return step_class(self)



"""YAML Data Provider"""

import yaml

from marvin.data.file_data_provider import FileDataProvider


class YAMLDataProvider(FileDataProvider):
    """
    DataDriven data provider for YAML/JSON sources.
    The data_source_id of this provider is the path to:
     * a YAML file with the following structure:
        ---
        setup:
            any: data
        iterations:
            - some: data
              for_iteration: 1
            - another: iteration
              with: more data
        tear_down:
            data_for: tear down

     * or a JSON file with the equivalent representation:
        {
          "setup": {"any": "data"},
          "iterations": [
            {"some": "data", "for_iteration": 1},
            {"another": "iteration", "with": "more data"},
          ],
          "tear_down": {"data_for": "tear down"}
        }
    """

    @classmethod
    def load_obj(cls, file_handle):
        return yaml.load(file_handle)

    @classmethod
    def supported_extensions(cls):
        return ['yml', 'yaml', 'json']



"""Null Data Provider"""
from marvin.data import DataProvider, IterationData


class NullDataProvider(DataProvider):
    """
    Default DataDriven data provider that yields a single iteration with no data.
    For cases where test scripts aren't implemented as a data-driven (more like unit test frameworks)
    """

    @classmethod
    def handles(cls, source_id):
        return source_id is None

    def __init__(self, *_args):
        super(NullDataProvider, self).__init__(None)

    @property
    def name(self):
        return None

    @property
    def description(self):
        return None

    @property
    def tags(self):
        return None

    @property
    def setup_data(self):
        return None

    @property
    def iterations(self):
        return [IterationData()]

    @property
    def tear_down_data(self):
        return None



from marvin.data.data_provider import DataProvider
from marvin.data.iteration_data import IterationData
from marvin.data.data_provider_registry import DataProviderRegistry
from marvin.data.file_data_provider import FileDataProvider
from marvin.data.data_providers.yaml_data_provider import YAMLDataProvider
from marvin.data.data_providers.null_data_provider import NullDataProvider

__all__ = ['DataProvider', 'FileDataProvider', 'DataProviderRegistry', 'IterationData']

DataProviderRegistry.register(YAMLDataProvider, NullDataProvider)



"""Marvin's Exceptions"""


class MarvinException(Exception):
    """Base exception for all marvin's exceptions"""
    def __init__(self, message=None):
        message = message or "An exception was raised in Marvin"
        super(MarvinException, self).__init__(message)


class ContextSkippedException(MarvinException):
    """Raised when a context entity (e.g. Step, TestScript, or Suite) is flagged to be skipped"""
    def __init__(self, context, reason=None):
        self.reason = reason or "Step skipped"
        self.context = context
        super(ContextSkippedException, self).__init__(self.reason)


class ExpectedExceptionNotRaised(MarvinException):
    """Raised when a step has defined exception expectations but nothing was raised"""
    def __init__(self, expectations):
        self.expectations = expectations
        expecting = expectations[0] if len(expectations) == 1 else expectations
        message = "No exception was raised, but the expectation was: %s" % expecting
        super(ExpectedExceptionNotRaised, self).__init__(message)


class StepsFailedInContext(MarvinException):
    """
    One or more steps have failed in this context
    This is not raised but used as a reason for parent context failure
    """
    def __init__(self, failed_steps):
        if failed_steps == 1:
            msg = "1 step has failed"
        else:
            msg = "{} steps have failed".format(failed_steps)
        super(StepsFailedInContext, self).__init__(msg)



from marvin import TestScript

from steps.check_website_title import CheckWebsiteTitle


class WebsiteTitles(TestScript):
    """Checks website titles"""
    NAME = "Website Titles"
    TAGS = ["web", "title"]

    def setup(self, data):
        pass

    def run(self, data):
        self.step(CheckWebsiteTitle).execute(data["website"], data["expected_title"])

    def tear_down(self, _data):
        # Clean up / logout, etc
        pass



import time
import json
from datetime import datetime

from marvin.report import EventType
from marvin.util.compat import urllib_mod


class ELKReporter(object):

    def __init__(self, publisher, base_url, index, schema_type="marvin"):
        publisher.subscribe(self.on_step_ended, EventType.STEP_ENDED)
        publisher.subscribe(self.on_test_started, EventType.TEST_STARTED)
        publisher.subscribe(self.on_test_ended, EventType.TEST_ENDED)
        publisher.subscribe(self.on_suite_ended, EventType.SUITE_ENDED)
        self.tests = []
        self._current_test = None
        self._base_url = base_url
        self._index = index
        self._type = schema_type

    def on_test_started(self, event):
        test_script = event.test_script

        self._current_test = {
            "name": test_script.name,
            "description": test_script.description,
            "tags": test_script.tags,
            "steps": []
        }

    def on_test_ended(self, event):
        self._current_test["status"] = event.status
        self._current_test["timestamp"] = self._ms_to_iso(event.timestamp)
        self._current_test["duration"] = event.duration
        self.tests.append(self._current_test)

    def on_step_ended(self, event):
        step = event.step
        step_info = {
            "name": step.name,
            "description": step.description,
            "tags": step.tags,
            "status": event.status,
            "timestamp": self._ms_to_iso(event.timestamp),
            "duration": event.duration
        }
        self._current_test['steps'].append(step_info)

    def on_suite_ended(self, _event):
        self._publish()

    def _publish(self):
        url = self._base_url + self._index + '/' + self._type
        data = json.dumps({"tests": self.tests, "timestamp": self._ms_to_iso(time.time() * 1000)})
        urllib = urllib_mod()
        req = urllib.Request(url, data, {'Content-Type': 'application/json'})
        urllib.urlopen(req)

    def _ms_to_iso(self, milliseconds):
        return datetime.fromtimestamp(milliseconds/1000.0).isoformat()



import sys
import traceback

from colorama import Fore

import marvin.util.files
from marvin.core.status import Status
from marvin.report import EventType as E
from marvin.exceptions import ContextSkippedException


COLORS = {
    Status.PASS: Fore.GREEN + '%s' + Fore.RESET,
    Status.FAIL: Fore.RED + '%s' + Fore.RESET,
    Status.SKIP: Fore.YELLOW + '%s' + Fore.RESET,
    'HEADER': Fore.CYAN + '%s' + Fore.RESET,
    'PHASE': Fore.MAGENTA + '%s' + Fore.RESET
}

INDENT = "  "

PHASE_TITLES = {
    E.TEST_SETUP_STARTED: "SETUP",
    E.TEST_SETUP_ENDED: "SETUP",
    E.TEST_ITERATION_STARTED: "ITERATION",
    E.TEST_ITERATION_ENDED: "ITERATION",
    E.TEST_TEARDOWN_STARTED: "TEARDOWN",
    E.TEST_TEARDOWN_ENDED: "TEARDOWN"
}


class EventLogger(object):
    """
    Prints the summary of test execution results in a terminal or file
    """

    def __init__(self, publisher, dest=sys.stdout):
        """
        Builds and subscribes an EventLogger
        :param publisher: The context's publisher instance
        :param dest: The output destination (defaults to sys.stdout)
        """
        publisher.subscribe(self.on_step_started, E.STEP_STARTED)
        publisher.subscribe(self.on_step_ended, E.STEP_ENDED)
        publisher.subscribe(self.on_test_started, E.TEST_STARTED)
        publisher.subscribe(self.on_test_ended, E.TEST_ENDED)
        publisher.subscribe(self.on_phase_started,
                            E.TEST_SETUP_STARTED,
                            E.TEST_ITERATION_STARTED,
                            E.TEST_TEARDOWN_STARTED)
        publisher.subscribe(self.on_phase_ended,
                            E.TEST_SETUP_ENDED,
                            E.TEST_ITERATION_ENDED,
                            E.TEST_TEARDOWN_ENDED)
        publisher.subscribe(self.collect_iteration_status, E.TEST_ITERATION_ENDED)
        publisher.subscribe(self.on_suite_ended, E.SUITE_ENDED)

        self._o = dest
        self._color = marvin.util.files.supports_color(self._o)
        self._suite_status = []
        self._reset_test_summary()

    def collect_iteration_status(self, event):
        self._current_test["iterations"][event.status] += 1

    def on_phase_started(self, event):
        title = PHASE_TITLES[event.event_type]
        phase_header = self._in_color('PHASE', title)
        phase_suffix = self._get_phase_suffix(event)
        self._p("%s[%s]%s", INDENT, phase_header, phase_suffix)

    def on_phase_ended(self, event):
        title = PHASE_TITLES[event.event_type]
        phase_header = self._in_color('PHASE', title)
        phase_suffix = self._get_phase_suffix(event)
        status = self._colored_status(event.status)
        self._p("%s[%s - %s]%s (%s)", INDENT, phase_header, status, phase_suffix, self._format_duration(event.duration))

    def on_step_started(self, event):
        step = event.step
        indent = INDENT * step.level
        self._p("%s%s: %s", indent, step.name, step.description)

    def on_step_ended(self, event):
        step = event.step
        status = self._colored_status(event.status)
        indent = INDENT * step.level
        self._p("%s[%s] %s (%s)", indent, status, step.name, self._format_duration(event.duration))

        if event.status == Status.SKIP and event.exception[0] == ContextSkippedException:
            self._p("%sSkip reason: %s", indent, event.exception[1].reason)
        elif event.status != Status.PASS and event.exception[2]:
            self._p("%s", self._format_exception(event.exception))

    def on_test_started(self, event):
        test_script = event.test_script
        self._reset_test_summary(test_script.name)
        test_header = self._in_color('HEADER', 'TEST')
        self._p("-" * 64)
        self._p("[%s] %s - %s", test_header, test_script.name, test_script.description)

    def _format_exception(self, exc_info):
        return "%s\n%s" % (repr(exc_info[1]), ''.join(traceback.format_tb(exc_info[2])))

    def on_test_ended(self, event):
        test_script = event.test_script
        test_header = self._in_color('HEADER', 'TEST')
        status = self._colored_status(event.status)
        self._current_test["status"] = event.status

        self._suite_status.append(self._current_test)
        if event.status != Status.PASS:
            for exc_info in event.exceptions:
                self._current_test["exceptions"].append(self._format_exception(exc_info))
        self._p("[%s - %s] %s (%s)", test_header, status, test_script.name, self._format_duration(event.duration))

    def on_suite_ended(self, event):
        self._p("\n" + "-" * 64)
        self._p("Execution Summary")
        self._p("-" * 64 + "\n")
        for test in self._suite_status:
            status = self._colored_status(test['status'])
            ipass = test['iterations'][Status.PASS]
            ifail = test['iterations'][Status.FAIL]
            iskip = test['iterations'][Status.SKIP]
            iall = ipass + ifail + iskip
            iteration_summary = "%d iteration(s) (%d pass - %d fail - %d skip)" % (iall, ipass, ifail, iskip)
            self._p("[%s] %s: %s", status, test['name'], iteration_summary)
            [self._p("%s", exception) for exception in test['exceptions']]

    def _get_phase_suffix(self, event):
        suffix = ""
        if event.event_type in (E.TEST_ITERATION_STARTED, E.TEST_ITERATION_ENDED) and event.iteration.name:
            suffix += ' ' + event.iteration.name
        if event.event_type == E.TEST_ITERATION_STARTED and event.iteration.description:
            suffix += ': ' + event.iteration.description
        return suffix

    def _reset_test_summary(self, test_name=None):
        self._current_test = {
            "name": test_name,
            "status": None,
            "exceptions": [],
            "iterations": {
                Status.PASS: 0,
                Status.FAIL: 0,
                Status.SKIP: 0
            }
        }

    def _p(self, s, *args):
        self._o.write(s % args + "\n")

    def _colored_status(self, status):
        return self._in_color(status, status)

    def _format_duration(self, millis):
        if millis < 1000:
            return "%d ms" % millis
        return "%.2f s" % (millis/1000.0)

    def _in_color(self, color, s):
        if self._color:
            return COLORS.get(color, "%s") % s
        return s



class FakeCode(object):
    def __init__(self, co_filename, co_name):
        self.co_filename = co_filename
        self.co_name = co_name


class FakeFrame(object):
    def __init__(self, f_code, f_globals):
        self.f_code = f_code
        self.f_globals = f_globals


class FakeTraceback(object):
    def __init__(self, frames, line_nums):
        if len(frames) != len(line_nums):
            raise ValueError("Ya messed up!")
        self._frames = frames
        self._line_nums = line_nums
        self.tb_frame = frames[0]
        self.tb_lineno = line_nums[0]

    @property
    def tb_next(self):
        if len(self._frames) > 1:
            return FakeTraceback(self._frames[1:], self._line_nums[1:])



import os
from marvin.data import DataProvider, IterationData


class FileDataProvider(DataProvider):
    """
    Base abstract class for file system based data providers
    """

    def __init__(self, *args, **kargs):
        super(FileDataProvider, self).__init__(*args, **kargs)
        with open(self._source_id, "r") as file_handle:
            self._data = self.load_obj(file_handle) or {}

    @classmethod
    def supported_extensions(cls):
        raise NotImplementedError("Method must be redefined in %s" % cls.__name__)

    @classmethod
    def load_obj(cls, file_handle):
        raise NotImplementedError("Method must be redefined in %s" % cls.__name__)

    @classmethod
    def handles(cls, source_id):
        if not source_id:
            return False
        _, ext = os.path.splitext(source_id)
        return ext and ext[1:].lower() in cls.supported_extensions()

    # Overrides
    @property
    def name(self):
        return self._data.get('name')

    # Overrides
    @property
    def description(self):
        return self._data.get('description')

    # Overrides
    @property
    def tags(self):
        tags = self._data.get('tags')
        return set(tags) if tags else None

    # Overrides
    @property
    def setup_data(self):
        return self._data.get('setup_data')

    # Overrides
    @property
    def iterations(self):
        for it_data in self._data.get('iterations', []):
            yield IterationData(**it_data)

    # Overrides
    @property
    def tear_down_data(self):
        return self._data.get('tear_down_data')



from tests.stubs.dummies import DummyData, DummyTest, DummyStep, DummyContext, DummyObserver, DummySuite, SampleStep
from tests.stubs.builders import IterationDataBuilder, TracebackBuilder

__all__ = ['DummyContext', 'DummyData', 'DummyObserver', 'DummyStep', 'DummySuite', 'DummyTest', 'IterationDataBuilder',
           'SampleStep', 'TracebackBuilder']




class DataProvider(object):
    """
    Interface to be implemented by any data provider.
    """

    def __init__(self, source_id):
        """
        All data provider's should support instantiation from one given argument representing
        the source of the data to be fetched.
        E.g. for file-based data source (e.g. JSon, XML, YML, etc) the data source id might be
        the path of the file. For a DB provider, it might be a record id
        """
        self._source_id = source_id

    @classmethod
    def handles(cls, source_id):
        """Returns true if this data provider supports loading the give data source id"""
        raise NotImplementedError("Method must be redefined in %s" % cls.__name__)

    @property
    def name(self):
        """
        Overrides the test case name
        :return: [None|String]
        """
        raise NotImplementedError("Method must be redefined in %s" % self.__class__.__name__)

    @property
    def description(self):
        """
        Overrides the test case description
        :return: [None|String]
        """
        raise NotImplementedError("Method must be redefined in %s" % self.__class__.__name__)

    @property
    def tags(self):
        """
        Extends the test case tag set
        :return: [None|iterable[String]]
        """
        raise NotImplementedError("Method must be redefined in %s" % self.__class__.__name__)

    @property
    def setup_data(self):
        """Returns a python object with the data to be passed to the 'setup' phase of the Test Script"""
        raise NotImplementedError("Method must be redefined in %s" % self.__class__.__name__)

    @property
    def iterations(self):
        """
        Returns python object's as a list or generator for each of the iterations
        Each yielded item is an instance of marvin.data.IterationData:
        """
        raise NotImplementedError("Method must be redefined in %s" % self.__class__.__name__)

    @property
    def tear_down_data(self):
        """Returns a python object with the data to be passed to the 'tear_down' phase of the Test Script"""
        raise NotImplementedError("Method must be redefined %s" % self.__class__.__name__)



from marvin.data import IterationData

from tests.stubs.exception import FakeTraceback, FakeFrame, FakeCode


class IterationDataBuilder(object):
    def __init__(self):
        self._args = {}

    def with_data(self, **kwargs):
        self._args['data'] = kwargs
        return self

    def with_name(self, name):
        self._args['name'] = name
        return self

    def with_description(self, description):
        self._args['description'] = description
        return self

    def with_tags(self, *args):
        self._args['tags'] = args
        return self

    def build(self):
        return IterationData(**self._args)


class TracebackBuilder(object):
    def __init__(self, frames=None):
        self._frames = frames or []

    def with_frame(self, filename, line, line_no):
        frame = {'filename': filename, 'line': line, 'line_no': line_no}
        frames = self._frames[:]
        frames.append(frame)
        return TracebackBuilder(frames)

    def build(self):
        frames = []
        line_nos = []
        for entry in self._frames:
            code = FakeCode(entry['filename'], entry['line'])
            frame = FakeFrame(code, {})
            frames.append(frame)
            line_nos.append(entry['line_no'])
        return FakeTraceback(frames, line_nos)



from marvin import Step, Suite, TestScript
from marvin.core.context import Context
from marvin.data import DataProvider, IterationData
from marvin.core.step_running_context import StepRunningContext
from marvin.core.test_running_context import TestRunningContext


class DummyContext(Context, StepRunningContext, TestRunningContext):

    def observer(self, *event_types):
        return DummyObserver(self.publisher, *event_types)

    def sub_context_finished(self, status):
        super(DummyContext, self).sub_context_finished(status)
        self.last_reported_status = status


class DummyObserver(object):

    def __init__(self, publisher, *event_types):
        self.events = []
        self.hooks = {}
        publisher.subscribe(self.on_event, *event_types)

    def hook(self, fn, *event_types):
        for event_type in event_types:
            self.hooks[event_type] = fn

    def on_event(self, event):
        self.events.append(event)
        if event.event_type in self.hooks:
            self.hooks[event.event_type](event)

    @property
    def last_event(self):
        return self.events[-1]


class DummySuite(Suite):

    def __init__(self):
        super(DummySuite, self).__init__()
        self._tests = []

    def add_test(self, test, data=None):
        self._tests.append((test, data))

    def tests(self):
        for t in self._tests:
            yield t

    def observer(self, *event_types):
        return DummyObserver(self.publisher, *event_types)


class DummyTest(TestScript):
    """Dummy Test used for testing"""

    def setup(self, data):
        self._do_something(data)

    def run(self, data):
        self._do_something(data)

    def tear_down(self, data):
        self._do_something(data)

    def _do_something(self, data):
        data = data or {}
        if 'fail' in data:
            raise Exception(data['fail'])
        if 'skip' in data:
            self.skip(data['skip'])


class SampleStep(Step):
    """Dummy Step used for testing"""

    def run(self, number_1, number_2, operation=int.__add__):
        return operation(number_1, number_2)


class DummyStep(Step):
    """Another Dummy Step"""
    def run(self, *args, **kwargs):
        if 'action' in kwargs:
            return kwargs.pop('action')(self, *args, **kwargs)
        elif 'skip' in kwargs:
            self.skip(kwargs['skip'])
        elif 'runtime_error' in kwargs:
            raise RuntimeError(kwargs['runtime_error'])
        return args, kwargs


class DummyData(DataProvider):
    """In-memory data provider that can be built dynamically"""
    def __init__(self):
        super(DummyData, self).__init__(None)
        self._data = {
            'name': None,
            'description': None,
            'tags': None,
            'setup_data': {},
            'iterations': [],
            'tear_down_data': {}
        }

    @property
    def name(self):
        return self._data['name']

    @property
    def description(self):
        return self._data['description']

    @property
    def tags(self):
        return self._data['tags']

    # Override
    @property
    def setup_data(self):
        return self._data['setup_data']

    # Override
    @property
    def iterations(self):
        return self._data['iterations'] or [IterationData()]

    # Override
    @property
    def tear_down_data(self):
        return self._data['tear_down_data']

    def with_name(self, name):
        self._data['name'] = name
        return self

    def with_description(self, description):
        self._data['description'] = description
        return self

    def wih_tags(self, *args):
        self._data['tags'] = set(args)
        return self

    def with_setup_data(self, **kwargs):
        self._data['setup_data'] = kwargs
        return self

    def with_iteration(self, iteration):
        self._data['iterations'].append(iteration)
        return self

    def with_tear_down_data(self, **kwargs):
        self._data['tear_down_data'] = kwargs
        return self




class Publisher(object):

    def __init__(self):
        self._observers = {}

    def subscribe(self, observer, *event_types):
        for event_type in event_types:
            self._observers.setdefault(event_type, []).append(observer)

    def notify(self, event):
        for observer in self._observers.get(event.event_type, []):
            observer(event)




class DataProviderRegistry(object):
    """
    Keeps track of the available DataProvider classes
    """

    DATA_PROVIDERS = []

    @classmethod
    def register(cls, *data_provider_classes):
        for data_provider_class in data_provider_classes:
            if data_provider_class not in cls.DATA_PROVIDERS:
                cls.DATA_PROVIDERS.insert(0, data_provider_class)

    @classmethod
    def data_provider_for(cls, data_source_id):
        return next((data_provider_class(data_source_id)
                     for data_provider_class in cls.DATA_PROVIDERS
                     if data_provider_class.handles(data_source_id)), None)

    @classmethod
    def supported_file_extensions(cls):
        return set(ext
                   for dp in cls.DATA_PROVIDERS if hasattr(dp, 'supported_extensions')
                   for ext in dp.supported_extensions())



import re
from setuptools import setup, find_packages


with open('marvin/version.py', 'r') as fd:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
        fd.read(),
        re.MULTILINE
    ).group(1)


setup(name='marvin-test',
      version=version,
      description='Marvin - Automation Framework',
      url='https://github.com/argos83/marvin',
      author='Sebastian Tello',
      author_email='argos83@gmail.com',
      packages=find_packages(exclude=("tests", "examples")),
      include_package_data=True,
      install_requires=['colorama', 'pyyaml'],
      tests_require=['pytest', 'pytest-cov', 'flake8', 'freezegun'],
      zip_safe=False,
      entry_points={
          'console_scripts': [
              'marvin=marvin.runner.cli:main'
          ]
      },
      classifiers=[
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6'
      ]
)



from marvin.core.reportable import Reportable
from marvin.core.context import Context
from marvin.core.step_running_context import StepRunningContext
from marvin.core.step_runner import StepRunner


class Step(Context, StepRunningContext, Reportable):

    def __init__(self, parent_context):
        Context.__init__(self, parent_context=parent_context)
        StepRunningContext.__init__(self)
        Reportable.__init__(self)
        self._shall_pass = False
        self._safe_exec = False
        self._expected_exceptions = []

    @property
    def do_not_fail(self):
        """
        Flag this step as 'optional' and set the status to PASS even if it fails
        Notice that the step may still raise exceptions that you might want to catch
        or silently ignore them via `safely`.
        """
        self._shall_pass = True
        return self

    @property
    def safely(self):
        """
        Don't propagate exceptions raised by this step
        (the status will still be set to FAIL unless `do_not_fail` is used.
        """
        self._safe_exec = True
        return self

    def expect_exception(self, *specs):
        """
        Expect this step to raise any of the given exception expectations.
        Each expectation can be either:
            * a Exception class. E.g. RuntimeError
            * a callable receiving the exception instance as argument and returning
            a Truthy value if the exception was expected. E.g. lambda e: isintance(e, HTTPError) and e.code == 404

        If exception expectations have been defined and the step raises no exception,
        then a ExpectedExceptionNotRaised is raised and the step status set to FAIL
        """
        self._expected_exceptions.extend(specs)
        return self

    def execute(self, *args, **kwargs):
        """
        Executes the step
        """
        runner = self.do(*args, **kwargs)
        with runner:
            pass

        return runner.result

    def do(self, *args, **kwargs):
        return StepRunner(self, args, kwargs)

    @property
    def expected_exceptions(self):
        """
        INTERNAL USE
        :return: a list of all defined exception expectations
        """
        return self._expected_exceptions

    @property
    def shall_pass(self):
        """
        INTERNAL USE
        :return: True if `do_no_fail` was set, False otherwise
        """
        return self._shall_pass

    @property
    def safe_exec(self):
        """
        INTERNAL USE
        :return: True if `safely` was set, False otherwise
        """
        return self._safe_exec

    def run(self, *args, **kargs):
        """Step's logic implementation (to be redefined by sub-classes)"""
        raise NotImplementedError("Method run must be implemented in step '%s'" % self.name)



from marvin.core.context import Context


class TestRunningContext(object):
    """
    A Mix-In for contexts that are allowed to run tests.
    E.g. Suites can run tests
    """

    def __init__(self):
        assert isinstance(self, Context)

    def test(self, test_class):
        """Instantiates a new test with this as it's parent context"""
        # TODO: assert test_class is subclsas of TestScript
        return test_class(self)

    def tags_match(self, tags):
        """
        To be redefined by specific Suite implementations that want to support tests/data/iteration filtering by tags
        :param tags: [set] a set of tags
        :return: [boolean] True if the set of tags match the filter implementation, False otherwise
        """
        return True



import inspect
import os
import sys


def supports_color(file_obj):
    """
    Returns True if the running system's terminal supports color, and False otherwise.
    """
    plat = sys.platform
    supported_platform = plat != 'Pocket PC' and (plat != 'win32' or 'ANSICON' in os.environ)
    # isatty is not always implemented
    is_a_tty = hasattr(file_obj, 'isatty') and file_obj.isatty()
    return supported_platform and is_a_tty


class ClassLoader(object):
    """
    Inspects the given module looking for a class member matching the given specifications
    """
    def __init__(self, parent_class, ignore_classes=None):
        self._parent = parent_class
        self._ignore = ignore_classes or []

    def find(self, mod):
        for member in mod.__dict__.values():
            if self._matches(mod, member):
                return member

    def _matches(self, mod, member):
        return (inspect.isclass(member)
                and issubclass(member, self._parent)
                and member != self._parent
                and member not in self._ignore
                and member.__module__ == mod.__name__)


class FileFinder(object):
    """
    Recursively walks a directory looking for files that satisfy the given filter
    """
    def __init__(self, search_path, filename_matcher=None):
        self._filename_matcher = filename_matcher
        self._search_path = search_path

    def find_all(self):
        for dirpath, dirnames, filenames in os.walk(self._search_path):
            for filename in filenames:
                if not self._filename_matcher or self._filename_matcher(filename):
                    yield os.path.join(dirpath, filename)

            blacklist = [dn for dn in dirnames if dn.startswith(".")]
            for dn in blacklist:
                dirnames.remove(dn)



import keyword
import os.path
import yaml


class Config(object):

    def __init__(self):
        self._data = {}

    def load(self, *config_files):
        for config_file in config_files:
            self.load_into(None, config_file)

    def load_into(self, namespace, config_file):
        name, ext = os.path.splitext(os.path.basename(config_file))
        with open(config_file) as fh:
            value = self._load_file(fh, ext.lower())
        self.set(namespace or name, value)

    def _load_file(self, fh, file_type):
        if file_type in ['.yaml', '.yml', '.json']:
            return yaml.load(fh)
        else:
            raise ValueError("Unsupported config extension: '%s'" % file_type)

    def set(self, name, obj):
        if keyword.iskeyword(name):
            raise ValueError("Can't use reserved name '%s' as config item" % name)
        self._data[name] = self._deep_merge(self._data.get(name), obj)

    def _deep_merge(self, base, update):
        if not (isinstance(base, dict) and isinstance(update, dict)):
            return update

        for key in update:
            if key in base and base[key] != update[key]:
                base[key] = self._deep_merge(base[key], update[key])
            else:
                base[key] = update[key]
        return base

    def __getattr__(self, name):
        if name in self._data:
            return self._data[name]
        raise AttributeError("Config setting '%s' not set" % name)



import time

"""Events that occur in Marvin"""


class EventType(object):
    SUITE_STARTED = 10
    SUITE_ENDED = 20
    TEST_STARTED = 30
    TEST_ENDED = 40
    TEST_SETUP_STARTED = 50
    TEST_SETUP_ENDED = 60
    TEST_ITERATION_STARTED = 70
    TEST_ITERATION_ENDED = 80
    TEST_TEARDOWN_STARTED = 90
    TEST_TEARDOWN_ENDED = 100
    STEP_STARTED = 110
    STEP_ENDED = 120
    STEP_SKIPPED = 130


class Event(object):
    """Abstract root Event class"""
    def __init__(self):
        self._timestamp = int(time.time() * 1000)

    @property
    def timestamp(self):
        """Event timestamp (Unix time with ms granularity)"""
        return self._timestamp


# -- SUITE related events --


class SuiteEvent(Event):
    """Abstract class for all Suite related events"""

    def __init__(self, suite):
        super(SuiteEvent, self).__init__()
        self._suite = suite

    @property
    def suite(self):
        """The Suite instance associated with this event"""
        return self._suite


class SuiteStartedEvent(SuiteEvent):
    """Triggered when a suite is about to start it's execution"""
    event_type = EventType.SUITE_STARTED


class SuiteEndedEvent(SuiteEvent):
    """Triggered when a suite has finished it's execution"""
    event_type = EventType.SUITE_ENDED

    def __init__(self, suite, start_time, status):
        super(SuiteEndedEvent, self).__init__(suite)
        self._start_time = start_time
        self._status = status
        self._duration = self.timestamp - start_time

    @property
    def start_time(self):
        """The timestamp when the suite started"""
        return self._start_time

    @property
    def duration(self):
        """The suite execution time in ms"""
        return self._duration

    @property
    def status(self):
        """The suite's status"""
        return self._status


# -- TEST related events --


class TestEvent(Event):
    """Abstract class for all TestScript related events"""

    def __init__(self, test_script, data_provider):
        super(TestEvent, self).__init__()
        self._test_script = test_script
        self._data_provider = data_provider

    @property
    def test_script(self):
        """The TestScript instance associated with this event"""
        return self._test_script

    @property
    def data_provider(self):
        """The instance of the data provider driving this test execution"""
        return self._data_provider


class TestStartedEvent(TestEvent):
    """Triggered when a test is about to start it's execution"""
    event_type = EventType.TEST_STARTED


class TestEndedEvent(TestEvent):
    """Triggered when a test has finished it's execution"""
    event_type = EventType.TEST_ENDED

    def __init__(self, test_script, data_provider, start_time, status, exceptions):
        super(TestEndedEvent, self).__init__(test_script, data_provider)
        self._start_time = start_time
        self._status = status
        self._exceptions = exceptions
        self._duration = self.timestamp - start_time

    @property
    def start_time(self):
        """The timestamp when the test started"""
        return self._start_time

    @property
    def duration(self):
        """The test execution time in ms"""
        return self._duration

    @property
    def status(self):
        """The test's status"""
        return self._status

    @property
    def exceptions(self):
        """
        The exceptions raised during this test. A list of 3-tuple items (like sys.exc_info)
        composed of (Exception type, Exception instance, Traceback instance)
        """
        return self._exceptions


class TestPhaseStartedEvent(TestEvent):
    """Abstract class for test's phase started: setup, iteration(s), tear down"""

    def __init__(self, test_script, data_provider, data):
        super(TestPhaseStartedEvent, self).__init__(test_script, data_provider)
        self._data = data

    @property
    def data(self):
        """the data for this test phase"""
        return self._data


class TestPhaseEndedEvent(TestEvent):
    """Abstract class for test's phase ended: setup, iteration(s), tear down"""

    def __init__(self, test_script, data_provider, data, start_time, status, exception):
        super(TestPhaseEndedEvent, self).__init__(test_script, data_provider)
        self._data = data
        self._start_time = start_time
        self._status = status
        self._exception = exception
        self._duration = self.timestamp - start_time

    @property
    def data(self):
        """the data used by this test phase"""
        return self._data

    @property
    def start_time(self):
        """The timestamp when the phase started"""
        return self._start_time

    @property
    def duration(self):
        """The phase execution time in ms"""
        return self._duration

    @property
    def status(self):
        """The phase's status"""
        return self._status

    @property
    def exception(self):
        """
        The exception raised by this phase (if any) as a 3-tuple (like sys.exc_info)
        composed of (Exception type, Exception instance, Traceback instance)
        """
        return self._exception


class TestSetupStartedEvent(TestPhaseStartedEvent):
    """Triggered when a test's setup phase is about to start"""
    event_type = EventType.TEST_SETUP_STARTED


class TestSetupEndedEvent(TestPhaseEndedEvent):
    """Triggered when a test's setup phase has concluded"""
    event_type = EventType.TEST_SETUP_ENDED


class TestIterationStartedEvent(TestPhaseStartedEvent):
    """Triggered when a test's iteration is about to start"""
    event_type = EventType.TEST_ITERATION_STARTED

    def __init__(self, test_script, data_provider, iteration):
        super(TestIterationStartedEvent, self).__init__(test_script, data_provider, iteration.data)
        self._iteration = iteration

    @property
    def iteration(self):
        """The IterationData instance with info about this iteration"""
        return self._iteration


class TestIterationEndedEvent(TestPhaseEndedEvent):
    """Triggered when a test's iteration has concluded"""
    event_type = EventType.TEST_ITERATION_ENDED

    def __init__(self, test_script, data_provider, iteration, start_time, status, exception):
        super(TestIterationEndedEvent, self).__init__(test_script, data_provider, iteration.data,
                                                      start_time, status, exception)
        self._iteration = iteration

    @property
    def iteration(self):
        """The IterationData instance with info about this iteration"""
        return self._iteration


class TestTearDownStartedEvent(TestPhaseStartedEvent):
    """Triggered when a test's tear down phase is about to start"""
    event_type = EventType.TEST_TEARDOWN_STARTED


class TestTearDownEndedEvent(TestPhaseEndedEvent):
    """Triggered when a test's tear down phase has concluded"""
    event_type = EventType.TEST_TEARDOWN_ENDED


# -- STEP related events --


class StepEvent(Event):
    """Abstract class for all Step related events"""
    def __init__(self, step):
        super(StepEvent, self).__init__()
        self._step = step

    @property
    def step(self):
        """The step instance associated with this event"""
        return self._step


class StepStartedEvent(StepEvent):
    """Triggered when a Step is about to be executed"""
    event_type = EventType.STEP_STARTED

    def __init__(self, step, args, kwargs):
        super(StepStartedEvent, self).__init__(step)
        self._args = args
        self._kwargs = kwargs

    @property
    def args(self):
        """The arguments being passed to the step's run method"""
        return self._args

    @property
    def kwargs(self):
        """The keyword arguments being passed to the step's run method"""
        return self._kwargs


class StepEndedEvent(StepEvent):
    """Triggered when a Step has finished it's execution"""
    event_type = EventType.STEP_ENDED

    def __init__(self, step, status, result, start_time, exception=(None, None, None)):
        super(StepEndedEvent, self).__init__(step)
        self._status = status
        self._result = result
        self._start_time = start_time
        self._duration = self.timestamp - start_time
        self._exception = exception

    @property
    def status(self):
        """The step's status"""
        return self._status

    @property
    def result(self):
        """The step's returned result (wrapped in a Result instance)"""
        return self._result

    @property
    def start_time(self):
        """The timestamp when the step started"""
        return self._start_time

    @property
    def duration(self):
        """The step execution time in ms"""
        return self._duration

    @property
    def exception(self):
        """
        The exception raised by this step (if any) as a 3-tuple (like sys.exc_info)
        composed of (Exception type, Exception instance, Traceback instance)
        """
        return self._exception


class StepSkippedEvent(StepEvent):
    """Triggered when a step is being skipped"""
    event_type = EventType.STEP_SKIPPED

    def __init__(self, step, exception):
        super(StepSkippedEvent, self).__init__(step)
        self._exception = exception

    @property
    def exception(self):
        """
        The corresponding ContextSkippedException containing the reason and stacktrace
        as a 3-tuple (like sys.exc_info) composed of (Exception type, Exception isntance, Traceback instance)
        """
        return self._exception



from marvin.core.reportable import Reportable
from marvin.core.context import Context
from marvin.core.step_running_context import StepRunningContext
from marvin.core.test_runner import TestRunner


class TestScript(Context, StepRunningContext, Reportable):
    """
    Base TestScript class to be extended by specific tests.
    Subclasses must re-implement the 'run' method, and can re-implement (if needed)
    the 'setup' and 'tear_down' methods.

    The execution of a test consist of:
        1. Execution of the 'setup' phase
        2. One or more executions of the 'run' phase
        3. Execution of the 'tear_down' phase

    Each of these methods will be called passing as argument an object representing data
    for the phase (usually this will be a python dictionary). The data is obtained from a
    DataProvider object, which should be capable of providing data for:
        * The 'setup' phase.
        * One or more iteration data for the 'run' phase (so it's up to the data provider to
            define how many iterations there will be)
        * The 'tear_down' phase.
    """

    def __init__(self, parent_context):
        Context.__init__(self, parent_context=parent_context)
        StepRunningContext.__init__(self)
        Reportable.__init__(self)
        self._failed_steps = 0

    def setup(self, _data):
        """
        Setup phase of a TestScript, executed only once, before all the iterations
        Can be redefined by subclasses implementing actual tests
        """
        pass

    def run(self, _data):
        """
        Main execution phase, invoked once per each iteration.
        Must be redefined by subclasses
        """
        raise NotImplementedError("Method run must be implemented in test script '%s'" % self.name)

    def tear_down(self, _data):
        """
        Tear down phase of a TestScript, executed only once, after all the iterations (even if there're failures)
        Can be redefined by subclasses implementingactual tests
        """
        pass

    def execute(self, data_provider=None):
        """
        Executes this TestScript with the given data provider
        """
        # TODO: TestScript instances should probably not be allowed to execute more than once
        # Either make them stateless (and allow re-execution) or add controls to fail if executed more than once
        TestRunner(self, data_provider).execute()
